#include <stdio.h>
#include <zephyr/kernel.h>
#include <zephyr/audio/dmic.h>
#include <zephyr/logging/log.h>

LOG_MODULE_REGISTER(dmic_sample);

#define MAX_SAMPLE_RATE  16000
#define SAMPLE_BIT_WIDTH 16
#define BYTES_PER_SAMPLE sizeof(int16_t)
#define READ_TIMEOUT     1000

#define BLOCK_SIZE(_sample_rate, _number_of_channels) \ 
    (BYTES_PER_SAMPLE * (_sample_rate / 10) * _number_of_channels)

#define MAX_BLOCK_SIZE   BLOCK_SIZE(MAX_SAMPLE_RATE, 2)
#define BLOCK_COUNT      10
K_MEM_SLAB_DEFINE_STATIC(mem_slab, MAX_BLOCK_SIZE, BLOCK_COUNT, 4);

// Define a buffer to store audio data
#define MAX_AUDIO_DATA_SIZE (MAX_BLOCK_SIZE * BLOCK_COUNT * 2) // Adjust the size as needed
uint8_t audio_data[MAX_AUDIO_DATA_SIZE];
size_t audio_data_index = 0;

static int do_pdm_transfer(const struct device *dmic_dev,
               struct dmic_cfg *cfg,
               size_t block_count)
{
    int ret;

    LOG_INF("PCM output rate: %u, channels: %u",
        cfg->streams[0].pcm_rate, cfg->channel.req_num_chan);

    ret = dmic_configure(dmic_dev, cfg);
    if (ret < 0) {
        LOG_ERR("Failed to configure the driver: %d", ret);
        return ret;
    }

    ret = dmic_trigger(dmic_dev, DMIC_TRIGGER_START);
    if (ret < 0) {
        LOG_ERR("START trigger failed: %d", ret);
        return ret;
    }

    for (int i = 0; i < block_count; ++i) {
        void *buffer;
        uint32_t size;

        ret = dmic_read(dmic_dev, 0, &buffer, &size, READ_TIMEOUT);
        if (ret < 0) {
            LOG_ERR("%d - read failed: %d", i, ret);
            // Stop the DMIC device before returning
            ret = dmic_trigger(dmic_dev, DMIC_TRIGGER_STOP);
            if (ret < 0) {
                LOG_ERR("STOP trigger failed: %d", ret);
            }
            return ret;
        }

        LOG_INF("%d - got buffer %p of %u bytes", i, buffer, size);
        // Print the received audio data
        LOG_HEXDUMP_INF((uint8_t *)buffer, size, "Received audio data:");

        // Store the received audio data into the buffer
        memcpy(&audio_data[audio_data_index], buffer, size);
        audio_data_index += size;

        k_mem_slab_free(&mem_slab, buffer);

        // Introduce a delay to ensure proper logging
        k_sleep(K_MSEC(100));
    }

    ret = dmic_trigger(dmic_dev, DMIC_TRIGGER_STOP);
    if (ret < 0) {
        LOG_ERR("STOP trigger failed: %d", ret);
        return ret;
    }

    return ret;
}

int main(void)
{
    const struct device *const dmic_dev = DEVICE_DT_GET(DT_NODELABEL(dmic_dev));
    int ret;

    LOG_INF("DMIC sample");

    if (!device_is_ready(dmic_dev)) {
        LOG_ERR("%s is not ready", dmic_dev->name);
        return 0;
    }

    struct pcm_stream_cfg stream = {
        .pcm_width = SAMPLE_BIT_WIDTH,
        .mem_slab  = &mem_slab,
    };
    struct dmic_cfg cfg = {
        .io = {
            .min_pdm_clk_freq = 1000000,
            .max_pdm_clk_freq = 3500000,
            .min_pdm_clk_dc   = 40,
            .max_pdm_clk_dc   = 60,
        },
        .streams = &stream,
        .channel = {
            .req_num_streams = 1,
        },
    };

    cfg.channel.req_num_chan = 1;
    cfg.channel.req_chan_map_lo =
        dmic_build_channel_map(0, 0, PDM_CHAN_LEFT);
    cfg.streams[0].pcm_rate = MAX_SAMPLE_RATE;
    cfg.streams[0].block_size =
        BLOCK_SIZE(cfg.streams[0].pcm_rate, cfg.channel.req_num_chan);

    ret = do_pdm_transfer(dmic_dev, &cfg, 2 * BLOCK_COUNT);
    if (ret < 0) {
        return 0;
    }

    // Process or save the audio data as needed
    // Example: Save the audio data to a file
    FILE *fp = fopen("D:/output/audio_data.bin", "wb"); 
    if (fp == NULL) 
    {
         LOG_ERR("Error opening the file.");
        return 1;
    }
    fwrite(audio_data, sizeof(uint8_t), audio_data_index, fp);
    fclose(fp);
    printf("Data has been written to audio_data.bin");

    LOG_INF("Exiting");
    return 0;
}